/*
 *  gui.c: User interface
 *
 *  Copyright (C) 2010- John Coppens (john@jcoppens.com)        
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *  
 *  You should have received a copy of the GNU Library General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <confmgr3.h>

#include "interface.h"
#include "support.h"
#include "main.h"
#include "pref.h"
#include "defs.h"
#include "gui.h"

GtkTextBuffer	*form_bff = NULL,
		*result_bff = NULL;
GtkTextView	*form_view = NULL,
		*result_view = NULL;
GtkTreeView	*conv_from_textview = NULL,
		*conv_to_textview = NULL,
		*variables_treeview = NULL;
GtkListStore	*conv_common_store = NULL,
		*variables_store = NULL;

int	dsp_mode = DSP_DECIMAL;
char *	mode_names[]   = {"Dec", "Hex", "Oct", "Bin"};
char *  mode_options[] = {"-d",  "-h",  "-o",  "-b"};

//	The following lists must be NULL terminated
char *	round_modes[] = {"none", "simple", "sig_fig", NULL};
char *	engin_modes[] = {"auto", "never",  "always",  NULL};


void
error_dialog(char *ttl, char *msg)
{
  GtkMessageDialog *dlg;

  dlg = GTK_MESSAGE_DIALOG(gtk_message_dialog_new(NULL,
                GTK_DIALOG_MODAL,
                GTK_MESSAGE_ERROR,
                GTK_BUTTONS_OK,  
                ttl, NULL)); 
  gtk_window_set_position(GTK_WINDOW(dlg), GTK_WIN_POS_CENTER);
  gtk_message_dialog_format_secondary_text(dlg, msg);
  gtk_dialog_run(GTK_DIALOG(dlg));
  gtk_widget_destroy(GTK_WIDGET(dlg));
}


void
insert_constant(char *wch)
{
  gtk_text_buffer_insert_at_cursor(form_bff, wch, -1);
}

void
insert_string(char *wch)
{
  gtk_text_buffer_insert_at_cursor(form_bff, wch, -1);
}

void
insert_function(char *intro, char *epilog)
{
  GtkTextIter b, e;
  GtkTextMark *endmark;
  int pos;

  if (gtk_text_buffer_get_selection_bounds(form_bff, &b, &e)) {	
    // if there is a selection, use it as argument for the function
    endmark = gtk_text_buffer_create_mark(form_bff, NULL, &e, TRUE);
    gtk_text_buffer_insert(form_bff, &b, intro, -1);
    gtk_text_buffer_get_iter_at_mark(form_bff, &e, endmark);
    gtk_text_buffer_insert(form_bff, &e, epilog, -1);
  } else {
    // Just insert the function at the cursor
    gtk_text_buffer_insert_at_cursor(form_bff, intro, -1);
    g_object_get(G_OBJECT(form_bff), "cursor_position", &pos, NULL);
    gtk_text_buffer_insert_at_cursor(form_bff, epilog, -1);
    gtk_text_buffer_get_iter_at_offset(form_bff, &b, pos);
    gtk_text_buffer_place_cursor(form_bff, &b);
  }
}

int
get_precision(void)
{
  GtkWidget *w;
  int val;

  w = lookup_widget(MainWindow, "main_precision_scale");
  val = (int)gtk_range_get_value(GTK_RANGE(w));
  return val;
}

void
conv_load_families(void)
{
  GtkWidget *w;
  GtkListStore *store;
  GtkTreeIter iter;
  int i;

  w = lookup_widget(ConversionsWindow, "conv_family_cbbox");
  store = gtk_list_store_new(1, G_TYPE_STRING);
  gtk_combo_box_set_model(GTK_COMBO_BOX(w), GTK_TREE_MODEL(store));

  for (i = 0; conversions[i].name; i++) {
    gtk_list_store_append(store, &iter);
    gtk_list_store_set(store, &iter,
		0, conversions[i].name,
		-1);
  }
  gtk_combo_box_set_active(GTK_COMBO_BOX(w), 0);
}

void
set_MainWindow_appearance(void)
{
  GtkWidget *w;

  if (pref.gen_mode == OP_SIMPLE ||
      pref.gen_mode == OP_NORMAL) {
    gtk_widget_hide(lookup_widget(MainWindow, "main_A_btn"));
    gtk_widget_hide(lookup_widget(MainWindow, "main_B_btn"));   
    gtk_widget_hide(lookup_widget(MainWindow, "main_C_btn"));   
    gtk_widget_hide(lookup_widget(MainWindow, "main_D_btn"));   
    gtk_widget_hide(lookup_widget(MainWindow, "main_E_btn"));   
    gtk_widget_hide(lookup_widget(MainWindow, "main_F_btn"));   
    gtk_widget_hide(lookup_widget(MainWindow, "main_mode_btn"));   
  } else {
    gtk_widget_show(lookup_widget(MainWindow, "main_A_btn"));   
    gtk_widget_show(lookup_widget(MainWindow, "main_B_btn"));   
    gtk_widget_show(lookup_widget(MainWindow, "main_C_btn"));   
    gtk_widget_show(lookup_widget(MainWindow, "main_D_btn"));   
    gtk_widget_show(lookup_widget(MainWindow, "main_E_btn"));   
    gtk_widget_show(lookup_widget(MainWindow, "main_F_btn"));   
    gtk_widget_show(lookup_widget(MainWindow, "main_mode_btn"));
  }
  if (pref.gen_mode == OP_SIMPLE) {
    gtk_widget_hide(lookup_widget(MainWindow, "main_kbd_table"));
    gtk_window_resize(GTK_WINDOW(MainWindow), 220, 100);
  } else {
    gtk_widget_show(lookup_widget(MainWindow, "main_kbd_table"));
  }

  gtk_window_set_decorated(GTK_WINDOW(MainWindow), pref.gen_showborders);
  gtk_window_set_keep_above(GTK_WINDOW(MainWindow), pref.gen_stayontop);

  gtk_widget_set(lookup_widget(MainWindow, "menubar1"), 
		"visible", pref.gen_showmenu,
		NULL);
}

void
conv_initialize_panels(void)
{
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *col;

  conv_common_store = gtk_list_store_new(
		CONV_NR_COLS,
                G_TYPE_STRING,          // The visible string
                G_TYPE_STRING);         // The command
  conv_from_textview = GTK_TREE_VIEW(lookup_widget(ConversionsWindow, "conv_from_treeview"));
  conv_to_textview = GTK_TREE_VIEW(lookup_widget(ConversionsWindow, "conv_to_treeview"));

  gtk_tree_view_set_model(conv_from_textview, GTK_TREE_MODEL(conv_common_store));
  gtk_tree_view_set_model(conv_to_textview, GTK_TREE_MODEL(conv_common_store));

  renderer = gtk_cell_renderer_text_new();
  gtk_object_set(GTK_OBJECT(renderer),
		"ypad", 0,
		NULL);
  col = gtk_tree_view_column_new_with_attributes(
                "Convert from", renderer,
                "text", 0,
                NULL);
  gtk_tree_view_append_column(conv_from_textview, col);

  renderer = gtk_cell_renderer_text_new();
  gtk_object_set(GTK_OBJECT(renderer),
		"ypad", 0,
		NULL);
  col = gtk_tree_view_column_new_with_attributes(
                "Convert to", renderer,
                "text", 0,
                NULL);
  gtk_tree_view_append_column(conv_to_textview, col);
}

void
initialize_variables_textview(void)
{
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *col;

  variables_store = gtk_list_store_new(
		VARIABLES_NR_COLS,
                G_TYPE_STRING,          // The variable's name
                G_TYPE_STRING,		// ... its value
		G_TYPE_STRING);         // ... and a comment

  variables_treeview = GTK_TREE_VIEW(lookup_widget(VariablesWindow, "var_treeview"));
  gtk_tree_view_set_model(variables_treeview, GTK_TREE_MODEL(variables_store));

  renderer = gtk_cell_renderer_text_new();
  gtk_object_set(GTK_OBJECT(renderer),
		"ypad", 0,
		NULL);
  col = gtk_tree_view_column_new_with_attributes(
                "Name", renderer,
                "text", VARIABLES_NAME,
                NULL);
  gtk_tree_view_append_column(variables_treeview, col);

  renderer = gtk_cell_renderer_text_new();
  gtk_object_set(GTK_OBJECT(renderer),
		"ypad", 0,
		NULL);
  col = gtk_tree_view_column_new_with_attributes(
                "Value", renderer,
                "text", VARIABLES_VALUE,
                NULL);
  gtk_tree_view_append_column(variables_treeview, col);

  renderer = gtk_cell_renderer_text_new();
  gtk_object_set(GTK_OBJECT(renderer),
		"ypad", 0,
		NULL);
  col = gtk_tree_view_column_new_with_attributes(
                "Comment", renderer,
                "text", VARIABLES_COMMENT,
                NULL);
  gtk_tree_view_append_column(variables_treeview, col);
}

void
add_variable(char *id, char *val, char *comment)
{
  GtkWidget *w;
  GtkListStore *store;
  GtkTreeIter iter;

  w = lookup_widget(VariablesWindow, "var_treeview");
  store = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(w)));
  gtk_list_store_append(store, &iter);
  if (id) gtk_list_store_set(store, &iter, VARIABLES_NAME, id, -1);
  if (val) gtk_list_store_set(store, &iter, VARIABLES_VALUE, val, -1);
  if (comment) gtk_list_store_set(store, &iter, VARIABLES_COMMENT, comment, -1);
}

void
pref_load_comboboxes(void)
{
  GtkWidget *w;

  w = lookup_widget(PrefWindow, "pref_dsp_engineering_cbbox");
  load_combo_box_with_strings(GTK_COMBO_BOX(w), engin_modes);

  w = lookup_widget(PrefWindow, "pref_dsp_rounding_cbbox");
  load_combo_box_with_strings(GTK_COMBO_BOX(w), round_modes);
}

//////	Extras popup
//	ExtrasPopup - functions
void
sine1_activate(void)
{
  insert_function("sin(", ")");
}

void
cosine1_activate(void)
{
  insert_function("cos(", ")");
}

void
tangent1_activate(void)
{
  insert_function("tan(", ")");
}

void
secant1_activate(void)
{
  insert_function("sec(", ")");
}

void
cosecant1_activate(void)
{
  insert_function("cosec(", ")");
}

void
cotangent1_activate(void)
{
  insert_function("cot(", ")");
}


void
arcsine1_activate(void)
{
  insert_function("asin(", ")");
}

void
arccosine1_activate(void)
{
  insert_function("acos(", ")");
}

void
arctangent1_activate(void)
{
  insert_function("atan(", ")");
}

void
arcsecant1_activate(void)
{
  insert_function("asec(", ")");
}

void
arccosecant1_activate(void)
{
  insert_function("acosec(", ")");
}

void
arccotangent1_activate(void)
{
  insert_function("acot(", ")");
}


void
hyperbolic_sine1_activate(void)
{
  insert_function("sinh(", ")");
}

void
hyperbolic_cosine1_activate(void)
{
  insert_function("cosh(", ")");
}

void
hyperbolic_tangent1_activate(void)
{
  insert_function("tanh(", ")");
}

void
hyperbolic_secant1_activate(void)
{
  insert_function("sech(", ")");
}

void
hyperbolic_cosecant1_activate(void)
{
  insert_function("cosech(", ")");
}

void
hyperbolic_cotangent1_activate(void)
{
  insert_function("coth(", ")");
}


void
inverted_hyperbolic_sine1_activate(void)
{
  insert_function("asinh(", ")");
}

void
inverted_hyperbolic_cosine1_activate(void)
{
  insert_function("acosh(", ")");
}

void
inverted_hyperbolic_tangent1_activate(void)
{
  insert_function("atanh(", ")");
}

void
inverted_hyperbolic_secant1_activate(void)
{
  insert_function("asech(", ")");
}

void
inverted_hyperbolic_cosecant1_activate(void)
{
  insert_function("acosech(", ")");
}

void
inverted_hyperbolic_cotangent1_activate(void)
{
  insert_function("atanh(", ")");
}

void
natural_logarithm1_activate(void)
{
  insert_function("ln(", ")");
}

void
base_10_logarithm1_activate(void)
{
  insert_function("log(", ")");
}

void
base_2_logarithm1_activate(void)
{
  insert_function("logtwo(", ")");
}

void
gamma_function1_activate(void)
{
  insert_function("Gamma(", ")");
}

void
log_gamma_function1_activate(void)
{
  insert_function("lnGamma(", ")");
}

void
zeta_function1_activate(void)
{
  insert_function("zeta(", ")");
}

void
sampling_function1_activate(void)
{
  insert_function("sinc(", ")");
}

void
nearest_integer1_activate(void)
{
  insert_function("round(", ")");
}

void
absolute_value1_activate(void)
{
  insert_function("abs(", ")");
}

void
ones_complement1_activate(void)
{
  insert_function("comp(", ")");
}

void
exponential_integral1_activate(void)
{
  insert_function("eint(", ")");
}

void
exponent1_activate(void)
{
  insert_function("exp(", ")");
}

void
factorial1_activate(void)
{
  insert_function("fact(", ")");
}

void
square_root1_activate(void)
{
  insert_function("sqrt(", ")");  
}

void
cube_root1_activate(void)
{
  insert_function("curt(", ")");
}

void
next_lower_integer1_activate(void)
{
  insert_function("floor(", ")");
}

void
next_higher_integer1_activate(void)
{
  insert_function("ceil(", ")");
}

void
random_value1_activate(void)
{
  insert_function("random(", ")");
}

void
random_integer1_activate(void)
{
  insert_function("irandom(", ")");
}


//      ExtrasPopup - Constants - General
void
pi1_activate(void)
{
  insert_constant(CONST_PI);
}

void
e1_activate(void)
{
  insert_constant(CONST_E);
}

void
eulers_constant1_activate(void)
{
  insert_constant(CONST_EULER);
}

void
catalan1_activate(void)
{
  insert_constant(CONST_CATALAN);
}

void
gravity1_activate(void)
{
  insert_constant(CONST_E_GRAVITY);
}

void
coulomb1_activate(void)
{
  insert_constant(CONST_COULOMB);
}

//      ExtrasPopup - Constants - Universal
void
impedancevacuum1_activate(void)
{
  insert_constant(CONST_Z_VACUUM);
}

void
epsilon1_activate(void)
{
  insert_constant(CONST_EPSILON0);
}

void
mu1_activate(void)
{
  insert_constant(CONST_MU0);
}

void
gravitational_constant1_activate(void)
{
  insert_constant(CONST_GRAVITY);
}

void
planck_constant1_activate(void)
{
  insert_constant(CONST_PLANCK);
}

void
speed_of_light1_activate(void)
{
  insert_constant(CONST_LIGHTSP);
}

//      ExtrasPopup - Constants - Atomic and Nuclear Constants
void
alpha_particle_mass1_activate(void)
{
  insert_constant(CONST_MALPHA);
}

void
bohr_radius1_activate(void)
{
  insert_constant(CONST_BOHR_RAD);
}

void
deuteron_mass1_activate(void)
{
  insert_constant(CONST_DEUT_MASS);
}

void
electron_mass1_activate(void)
{
  insert_constant(CONST_ELEC_MASS);
}

void
electron_radius1_activate(void)
{
  insert_constant(CONST_ELEC_RAD);
}

void
fermi_coupling_constant1_activate(void)
{
  insert_constant(CONST_FERMI);
}

void
fine_structure_constant1_activate(void)
{
  insert_constant(CONST_FINE_STR);
}

void
hartree_energy1_activate(void)
{
  insert_constant(CONST_HARTREE);
}

void
helion_mass1_activate(void)
{
  insert_constant(CONST_HELION);
}

void
muon_mass1_activate(void)
{
  insert_constant(CONST_MUON);
}

void
neutron_mass1_activate(void)
{
  insert_constant(CONST_NEUTRON);
}

void
proton_mass1_activate(void)
{
  insert_constant(CONST_PROTON);
}

void
rydberg_constant1_activate(void)
{
  insert_constant(CONST_RYDBERG);
}

void
tau_mass1_activate(void)
{
  insert_constant(CONST_TAU_MASS);
}

//	ExtrasPopup - Constants - Physio-chemical
void
atomic_mass_constant1_activate(void)
{
  insert_constant(CONST_ATOM_MASS);
}

void
avogadros_constant1_activate(void)
{
  insert_constant(CONST_AVOGADRO);
}

void
boltzmann_constant1_activate(void)
{
  insert_constant(CONST_BOLTZMANN);
}

void
faraday_constant1_activate(void)
{
  insert_constant(CONST_FARADAY);
}

void
first_radiation_constant1_activate(void)
{
  insert_constant(CONST_RAD_1);
}

void
loschmidt_constant1_activate(void)
{
  insert_constant(CONST_LOSCHMIDT);
}

void
molar_mass_constant1_activate(void)
{
  insert_constant(CONST_MOLAR_GAS);
}

void
molar_volumn_ideal_gas1_activate(void)
{
  insert_constant(CONST_MOLAR_VOL);
}

void
second_radiation_constant1_activate(void)
{
  insert_constant(CONST_RAD_2);
}

void
sigma1_activate(void)
{
  insert_constant(CONST_SIGMA);
}

void
wien_displacement_law1_activate(void)
{
  insert_constant(CONST_WIEN_DISPL);
}

void
bohr_magneton1_activate(void)
{
  insert_constant(CONST_BOHR_MAG);
}

void
nuclear_magneton1_activate(void)
{
  insert_constant(CONST_NUCL_MAG);
}

void
conductance_quantum1_activate(void)
{
  insert_constant(CONST_CONDUCT_Q);
}

void
elementary_charge1_activate(void)
{
  insert_constant(CONST_EL_CHARGE);
}

void
josephson_constant1_activate(void)
{
  insert_constant(CONST_JOSEPHSON);
}

void
von_klitzing_constant1_activate(void)
{
  insert_constant(CONST_KLITZING);
}

void
magnetic_flux_quantum1_activate(void)
{
  insert_constant(CONST_MAGFLUX_Q);
}

void
physio_chemical_constants1_activate(void)
{
  
}

void
a_random_value1_activate(void)
{
  insert_constant(CONST_RAND);
}

void
a_random_integer1_activate(void)
{
  insert_constant(CONST_IRAND);
}

void
quit2_activate(void)
{
  quit1_activate();
}

void
preferences2_activate(void)
{
  preferences1_activate();
}

void
variables2_activate(void)
{
  variables1_activate();
}

void
inspector1_activate(void)
{
  history1_activate();
}

//////	ResultPopup
void
save_to_variable3_activate(void)
{
  save_to_variable2_activate();
}

void
conversions3_activate(void)
{
  conversions2_activate();
}


//////  PrefWindow
void
pref_cancel_btn_clicked(void)
{
  gtk_widget_destroy(PrefWindow);
}

void
pref_ok_btn_clicked(void)
{
  parse_preferences();
  save_preferences();
  set_MainWindow_appearance();
  gtk_widget_destroy(PrefWindow);
}


void
PrefWindow_realize(void)
{
}

gboolean
PrefWindow_delete_event(void)
{
  return FALSE;
}


//////  MainWindow

//      MainWindow - menubar - Input
void
clear1_activate(void)
{
  clear_form_entry();
}

void
history1_activate(void)
{
  InspectorWindow = create_InspectorWindow();
  gtk_widget_show(InspectorWindow);
}

void
variables1_activate(void)
{
  VariablesWindow = create_VariablesWindow();
  gtk_widget_show(VariablesWindow);
}

void
preferences1_activate(void)
{
  PrefWindow = create_PrefWindow();
  pref_load_comboboxes();
  show_preferences();
  gtk_widget_show(PrefWindow);
}
  
void
quit1_activate(void)
{
  gtk_exit(0);
}

//      MainWindow - menubar - Output
void
save_to_variable2_activate(void)
{
  GtkWidget *w;
  GtkTextIter b, e;
  char *p;

  VariablesWindow = create_VariablesWindow();
  gtk_widget_show(VariablesWindow);

  AddVariableWindow = create_AddVariableWindow();
  w = lookup_widget(AddVariableWindow, "addvar_value_entry");
  gtk_text_buffer_get_bounds(result_bff, &b, &e);
  p = gtk_text_buffer_get_text(result_bff, &b, &e, FALSE);
  gtk_entry_set_text(GTK_ENTRY(w), p);
  g_free(p);

  gtk_widget_show(AddVariableWindow);
}

void
conversions2_activate(void)
{
  ConversionsWindow = create_ConversionsWindow();
  gtk_widget_show(ConversionsWindow);
}

void
decimal1_activate(void)
{
  set_display_mode(DSP_DECIMAL);
}

void
hexadecimal1_activate(void)
{
  set_display_mode(DSP_HEXADECIMAL);
}

void
octal1_activate(void)
{
  set_display_mode(DSP_OCTAL);
}

void
binary1_activate(void)
{
  set_display_mode(DSP_BINARY);
}

void
recalculate1_activate(void)
{
  execute_formula();
}

void
about1_activate(void)
{
}

//	MainWindow - events

gboolean
MainWindow_delete_event(void)
{
  gtk_exit(0);
  return FALSE;
}

void
MainWindow_realize(void)
{
  GtkWidget *w;
  GtkAdjustment *adj;
  PangoFontDescription *fdescr;

  fdescr = pango_font_description_from_string(pref.font_formula);
  form_view = GTK_TEXT_VIEW(lookup_widget(MainWindow, "main_input_txtview"));
  gtk_widget_modify_font(GTK_WIDGET(form_view), fdescr);
  form_bff = gtk_text_view_get_buffer(form_view);
  gtk_text_view_set_wrap_mode(form_view, GTK_WRAP_WORD_CHAR);
  pango_font_description_free(fdescr);

  fdescr = pango_font_description_from_string(pref.font_result);
  result_view = GTK_TEXT_VIEW(lookup_widget(MainWindow, "main_result_txtview"));
  gtk_widget_modify_font(GTK_WIDGET(result_view), fdescr);
  result_bff = gtk_text_view_get_buffer(result_view);
  gtk_text_view_set_wrap_mode(result_view, GTK_WRAP_WORD_CHAR);
  pango_font_description_free(fdescr);

  set_MainWindow_appearance();

  w = lookup_widget(MainWindow, "main_precision_scale");
  gtk_range_set_range(GTK_RANGE(w), 0, MAX_NR_DECIMALS);

  w = lookup_widget(MainWindow, "main_display_frame");
  gtk_widget_add_events(w, GDK_BUTTON_PRESS_MASK);
}

void
clear_form_entry(void)
{
  GtkTextIter b, e;

  gtk_text_buffer_get_bounds(form_bff, &b, &e);
  gtk_text_buffer_delete(form_bff, &b, &e);

  gtk_widget_grab_focus(GTK_WIDGET(form_view));
}

void
clear_result_entry(void)
{   
  GtkTextIter b, e;
 
  gtk_text_buffer_get_bounds(result_bff, &b, &e);
  gtk_text_buffer_delete(result_bff, &b, &e);
}
 

void
show_result(char *res)
{
  int w, h;
  char *p = N_(res);

  if (strstr(res, "@Inf@")) {
    p = N_("Infinity");
  } else
    if (strstr(res, "@NaN@")) {
      p = N_("Not a Number");
    }

  gtk_text_buffer_set_text(result_bff, _(p), -1);
  gtk_widget_get_size_request(MainWindow, &w, &h);
  if (pref.gen_mode != OP_SIMPLE)
    gtk_window_resize(GTK_WINDOW(MainWindow), 1, 1);
}

void
execute_formula(void)
{
  GtkTextIter b, e;
  int outfd[2], errfd[2];
  int pid_i, i;
  char	*p, *p1, *p2,
	outbff[STDOUT_BFF_SIZE],
	errbff[STDERR_BFF_SIZE];

  if (pipe(outfd) == -1) {
    fprintf(stderr, _("Creating pipe for stdout failed\n"));
    return;
  }

  if (pipe(errfd) == -1) {
    fprintf(stderr, _("Creating pipe for stderr failed\n"));
    return;
  }

  pid_i = fork(); 
  switch (pid_i) {
    case -1:
      fprintf(stderr, "Forking failed\n");
      return;

    case 0:	 			// Child: do wcalc
      gtk_text_buffer_get_bounds(form_bff, &b, &e);
      p = gtk_text_buffer_get_text(form_bff, &b, &e, FALSE);
      p1 = g_strdup_printf("-P%d", get_precision());
      p2 = g_strdup_printf("--bits%04d", pref.gen_intprecision);

      if (debug & DBG_COMMAND)
        printf("wcalc %s %s %s %s %s %s %s\n",
		p1,					// Precision
		mode_options[dsp_mode],			// Display mode (hex/dec/etc)
		pref.dsp_prefixes   ? "-p" : "",	// Use prefixes in output
		pref.exp_useradians ? "-r" : "",	// Use radians
		pref.gen_guard      ? "-c" : "",	// Conservative mode
		p2,					// Internal precision (bits)
		p					// Actual formula to evaluate
        );

      close(outfd[0]);				// Close read end of pipe
      dup2(outfd[1], STDOUT_FILENO);
      close(errfd[0]);				// Close read end of pipe
      dup2(errfd[1], STDERR_FILENO);

      execlp("wcalc", "wcalc",
		p1,					// Precision
		mode_options[dsp_mode],			// Display mode (hex/dec/etc)
		pref.dsp_prefixes   ? "-p" : "",	// Use prefixes in output
		pref.exp_useradians ? "-r" : "",	// Use radians
		pref.gen_guard      ? "-c" : "",	// Conservative mode
		p2,					// Internal precision (bits)
		p,					// Actual formula to evaluate
		(char *)NULL);

      // Should never get here...
      g_free(p); g_free(p1); g_free(p2);
      close(outfd[1]);
      close(errfd[1]);

    default:					// Parent
      close(outfd[1]);				// Close write end of stdout pipe
      close(errfd[1]);				// Close write end of stderr pipe
      waitpid(pid_i, 0, 0);       		// Wait for child
      read(outfd[0], outbff, sizeof(outbff));
      read(errfd[0], errbff, sizeof(errbff));
      close(outfd[0]);
      close(errfd[0]);

      if (debug & DBG_REPLY)
        printf("wcalc stdout reply: [%s]\n", outbff);
      if (debug & DBG_ERRORS)
        printf("wcalc stderr reply: [%s]\n", errbff);

      p = strchr(errbff, '\n');			// Remove trailing EOL
      if (p) *p = '\0';

      if (strstr(errbff, "Undefined variable:")) {
        error_dialog(_("Error"), _("Undefined variable"));
      }

      if (strstr(errbff, "Incorrect number format")) {
        if (strstr(errbff, "for octal")) {
          error_dialog(_("Error"), _("Incorrect number format - expected 0-7"));
        }
        if (strstr(errbff, "for binary")) {
          error_dialog(_("Error"), _("Incorrect number format - expected 0 or 1"));
        }
      }

      if (!(p = strchr(outbff, '\n')))  {	// EOL present?
        printf(_("Incomplete reply from wcalc (return missing)\n"));
        return;
      }
      *p = '\0';
      show_result(&outbff[3]);
  } // end switch
}

void
main_btn_clicked(int key)
{
  char bff[] = "?";

  switch (key) {
    case GDK_Num_Lock:		clear_form_entry(); break;
    case GDK_KP_Divide:		insert_string("/"); break;
    case GDK_KP_Multiply:	insert_string("*"); break;
    case GDK_KP_Add:		insert_string("+"); break;
    case GDK_KP_Subtract:	insert_string("-"); break;
    case GDK_KP_0:		insert_string("0"); break;
    case GDK_KP_1:		insert_string("1"); break;
    case GDK_KP_2:		insert_string("2"); break;
    case GDK_KP_3:		insert_string("3"); break;
    case GDK_KP_4:		insert_string("4"); break;
    case GDK_KP_5:		insert_string("5"); break;
    case GDK_KP_6:		insert_string("6"); break;
    case GDK_KP_7:		insert_string("7"); break;
    case GDK_KP_8:		insert_string("8"); break;
    case GDK_KP_9:		insert_string("9"); break;
    case GDK_A:			insert_string("a"); break;
    case GDK_B:			insert_string("b"); break;
    case GDK_C:			insert_string("c"); break;
    case GDK_D:			insert_string("d"); break;
    case GDK_E:			insert_string("e"); break;
    case GDK_F:			insert_string("f"); break;
    case GDK_KP_Decimal:	insert_string("."); break;
    case GDK_KP_Enter:		execute_formula(); break;
    default:			bff[0] = key;
				insert_string("A"); break;
  }
}

void
set_display_mode(int new_mode)
{
  GtkWidget *w;

  w = lookup_widget(MainWindow, "main_mode_btn");
  gtk_button_set_label(GTK_BUTTON(w), mode_names[new_mode]);

  if (new_mode == dsp_mode) return;
  dsp_mode = new_mode;
  clear_result_entry();
}

void
main_mode_btn_clicked(void)
{
  set_display_mode((dsp_mode + 1) % DSP_NR_MODES);
}

gboolean
main_input_txtview_button_press_event(GdkEventButton *event)
{
  if (event->button == 3) {
    ExtrasPopup = create_ExtrasPopup();
    gtk_menu_popup(GTK_MENU(ExtrasPopup), NULL, NULL, NULL, NULL,
		event->button, event->time);
    return TRUE;
  }
  return FALSE;
}

gboolean
main_input_txtview_key_press_event(GdkEventKey *event)
{
  if ((event->keyval == GDK_KP_Enter) ||
      (event->keyval == GDK_Return)) {
    execute_formula();
    return TRUE;
  }
  return FALSE;
}

gboolean
main_result_txtview_button_press_event(GdkEventButton *event)
{
  if (event->button == 3) {
    ResultPopup = create_ResultPopup();
    gtk_menu_popup(GTK_MENU(ResultPopup), NULL, NULL, NULL, NULL, event->button, event->time);
    return TRUE;
  }
  return FALSE;
}

gboolean
main_eventbox_button_press_event(GtkWidget *widget, GdkEventButton *event)
{
  if (event->button != 1)
    return FALSE;

  gtk_window_begin_move_drag(GTK_WINDOW(MainWindow), event->button, event->x_root,
		event->y_root, event->time); 
  return FALSE;
}


//////  ConversionsWindow
void
conv_family_cbbox_changed(void)
{
  GtkWidget *w;
  GtkTreeIter iter;
  int i, which;

  gtk_list_store_clear(conv_common_store);

  w = lookup_widget(ConversionsWindow, "conv_family_cbbox");
  which = gtk_combo_box_get_active(GTK_COMBO_BOX(w));

  for (i = 0; conversions[which].table[i]; i++) {
    gtk_list_store_append(conv_common_store, &iter);
    gtk_list_store_set(conv_common_store, &iter,
		0, conversions[which].table[i],
		-1);
  }
}

void
conv_from_treeview_cursor_changed(void)
{
}

void
conv_to_treeview_cursor_changed(void)
{
}

void
conv_cancel_btn_clicked(void)
{
  gtk_widget_destroy(ConversionsWindow);
}

void
conv_ok_btn_clicked(void)
{
  gtk_widget_destroy(ConversionsWindow);
}

gboolean
ConversionsWindow_delete_event(void)
{
  gtk_widget_destroy(ConversionsWindow);
  return FALSE;
}

void
ConversionsWindow_realize(void)
{
  conv_initialize_panels();
  conv_load_families();
}

//////  VariablesWindow
gboolean
VariablesWindow_delete_event(void)
{
  gtk_widget_destroy(VariablesWindow);
}

void
VariablesWindow_realize(void)
{
  initialize_variables_textview();
}

void
var_treeview_cursor_changed(void)
{
}

void
var_add_btn_clicked(void)
{
  AddVariableWindow = create_AddVariableWindow();
  gtk_widget_show(AddVariableWindow);
}

void
var_remove_btn_clicked(void)
{
  GtkWidget *w;
  GtkTreeSelection *sel;
  GtkListStore *store;
  GtkTreeIter iter;

  w = lookup_widget(VariablesWindow, "var_treeview");
  store = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(w)));
  sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(w));
  if (gtk_tree_selection_get_selected(sel, NULL, &iter)) {
    gtk_list_store_remove(store, &iter);
  }
}

void
var_clear_btn_clicked(void)
{
  GtkWidget *w;
  GtkListStore *store;
  GtkTreeIter iter;

  w = lookup_widget(VariablesWindow, "var_treeview");
  store = GTK_LIST_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(w)));
  gtk_list_store_clear(store);
}

void
var_cancel_btn_clicked(void)
{
}

void
var_ok_btn_clicked(void)
{
}


//////  InsprectorWindow
gboolean
InspectorWindow_delete_event(void)
{
  gtk_widget_destroy(InspectorWindow);
}

void
InspectorWindow_realize(void)
{
}

void
insp_which_cbbox_changed(void)
{
}

void
insp_treeview_cursor_changed(void)
{
}

void
insp_ok_btn_clicked(void)
{
}


//////  AddVariableWindow   

void
addvar_cancel_btn_clicked(void)
{
  gtk_widget_destroy(AddVariableWindow);
}

void
addvar_ok_btn_clicked(void)
{
  GtkWidget *w;
  char *id, *val, *com;

  w = lookup_widget(AddVariableWindow, "addvar_identifier_entry");
  id = (char *) gtk_entry_get_text(GTK_ENTRY(w));
  w = lookup_widget(AddVariableWindow, "addvar_value_entry");
  val = (char *) gtk_entry_get_text(GTK_ENTRY(w));
  w = lookup_widget(AddVariableWindow, "addvar_comment_entry");
  com = (char *) gtk_entry_get_text(GTK_ENTRY(w));

  add_variable(id, val, com);

  gtk_widget_destroy(AddVariableWindow);
}

void
AddVariableWindow_realize(void)
{

}

gboolean
AddVariableWindow_delete_event(void)
{
  return FALSE;
}
